#!/usr/bin/env bash
# @(#) GigA+ channel management script.

OSNAME=`uname -s`
: ${MAIL_AGENT:='mailx'}
hostnm=$(hostname)

#
# Functions:
#

errmsg()
{
  echo >&2 $@; exit 1
}


print_fmt_uptime()
{
    local asec=${1:?}

    local dy=$(( ${asec} / 86400 ))
    local dyl=$(( ${asec} % 86400 ))
    local hr=$(( ${dyl} / 3600 ))
    local hrl=$(( ${dyl} % 3600 ))
    local mi=$(( ${hrl} / 60 ))
    local sec=$(( ${hrl} % 60  ))

    printf "%02dd %02dh:%02dm:%02ds\n" $dy $hr $mi $sec
}

get_file_age()
{
    local fpath=${1:?}
    local last_mod_tm=

    case "$OSNAME" in
        'Linux') last_mod_tm=$(stat -c '%Y' $fpath) ;;
        'FreeBSD') eval $(stat -s $fpath); last_mod_tm=${st_mtime:?} ;;
        *) echo >&2 "$OSNAME not supported"; exit 1 ;;
    esac

    local now=$(date +"%s")
    local asec=$((${now:?} - ${last_mod_tm}))
    print_fmt_uptime $asec
}


get_channel_size()
{
    local duopt=${1:?}
    local croot=${2:?}
    local ctag=${3:?}
    local shards=${4}

    local csize=$(du ${duopt} ${croot}/${ctag} | cut -f1)
    if [ -n "${shards}" ]; then
        for s in $(echo ${shards} | tr ':' ' '); do
            local shdir=${croot}/${s:?}/${ctag}
            local shsize=$(du ${duopt} ${shdir} | cut -f1)
            if [ -n "${shsize}" ]; then
                let csize+=${shsize}
            fi
        done
    fi

    echo ${csize}
}


email_alert()
{
    local subj="$1"
    local head="$2"
    local rc=

    [ -z "${can_email}" ] && return 1
    [ -n "${cha_ADMIN_EMAIL}" ] || return 1

    local letter=${GXA_ROOT:?}/log/alert-email.$$
    echo "$head" >${letter}

    local vlog="${cha_ROOT:?}/${cha_ID:?}/vsm.log"
    if [ -f "$vlog" ]; then
        echo "==== ${vlog} ====" >>${letter}
        tail -50 ${vlog} >>${letter}
        echo "=================" >>${letter}
    else
        echo "No ${vlog} found, listing ${cha_ROOT:?}/${cha_ID:?}:" >>${letter}
        ls -l ${cha_ROOT:?}/${cha_ID:?} >>${letter}
    fi

    cat ${letter} | ${MAIL_AGENT} -s "${subj}" ${cha_ADMIN_EMAIL:?}; rc=$?
    [ "0" = "$rc" ] && rm -f ${letter}
}


unset_spec_env()
{
    unset cha_ID cha_GPM cha_URL cha_ROOT cha_ROLLOVER
    unset cha_TRANSOPT cha_SHARDS
    unset cha_CAPACITY cha_DURATION cha_MAX_CERR
    unset cha_FALSE_ROOT cha_GRAN_MASK cha_MITEMS
    unset cha_LMARK_DISKSIZE cha_HMARK_DISKSIZE
    unset cha_HOS cha_PRB cha_ADMIN_EMAIL
}


cha_status()
{
    local specfile=${1:?}
    local alert=
    local online="OFF"

    [ -f "${specfile}" ] || errmsg "Invalid spec: ${specfile}"
    unset_spec_env
    . ${specfile} >/dev/null

    local cdir=${cha_ROOT:?}/${cha_ID:?}
    if [ ! -d "${cdir}" ]; then
        echo >&2 "(ERR) ${cha_ID:?} - [$cdir] not found"
        return 1
    fi

    local csize=$(get_channel_size -Lsm ${cha_ROOT:?} ${cha_ID:?} ${cha_SHARDS})
    if [ -n "${cha_HMARK_DISKSIZE}" ]; then
        if [ "$csize" -ge "${cha_HMARK_DISKSIZE}" ]; then
            alert=1; aux="${aux} *OVSZ*"
        fi
    fi

    local pidfile=$cdir/vsm.pid
    if [ -f "$pidfile" ]; then
        local vsmpid=$(cat $pidfile)
        if ps -p ${vsmpid} 2>/dev/null >&2; then
            online="ON"
            aux="uptime: $(get_file_age ${pidfile})"
        else
            local old_pid=$(cat ${pidfile} 2>/dev/null)
            if rm -f ${pidfile}; then
                aux="- stale pidfile (${old_pid}) removed; "
            fi
        fi
    fi

    [ -f $cdir/REPLICA ] && online="REPL"
    [ -f $cdir/SUSPEND ] && online="SUSP"

    if [ "OFF" = "${online}" ]; then
        alert=1
    fi

    [ -n "${MUST_ALERT}" ] || alert=

    if [ -n "${alert}" ]; then
        aux="${aux} (ALERT!)"
    fi

    local report="(${online}) ${cha_ID:?} [${csize} MB] ${aux}"
    echo >&2 "$report"; aux=

    if [ -n "${alert}" ]; then
        email_alert "channel alert for [${cha_ID}] on ${hostnm}" "${report}"
        alert=
    fi

    [ "ON" = "${online}" ] && return 0
    return 1
}

get_status()
{
    local tag=${1:?}
    local cha=

    check_pidfile ${GPM_PIDFILE:?} gxpm

    echo >&2 "----- channels -----"
    if [ "all" = "$tag" ]; then
        find ${GXA_SPEC_DIR} -type f -name '*.spec' | \
            while read sname; do
                cha_status ${sname:?}
                cha=${cha_ID}
            done
    else
        for cha in `echo ${tag:?}`; do
            sname=${GXA_SPEC_DIR}/${cha:?}.spec
            if [ -f "${sname}" ]; then
                cha_status ${sname:?}
            else
                echo >&2 "No spec [${sname}] for ${cha}"
            fi
        done
    fi
    echo >&2 "--------------------"
}

check_pidfile()
{
    local hush=

    if [ '--hush' = "$1" ]; then
        hush=1; shift
    fi

    local fname=$1
    local app=${2:-'process'}
    local run_msg=${3:-'is running'}

    if [ -f "${fname}" ]; then
        pid=$(cat ${fname})
        [ -n "$pid" ] || return 1

        if ! ps -p ${pid} 2>/dev/null >&2; then
            [ -z "$hush" ] && echo "Stale pidfile [${fname}] found; removing."
            rm -f ${fname}
        fi
    fi

    if [ ! -f "${fname}" ]; then
        [ -z "$hush" ] && echo "[${app}] is not running."
        return 1
    fi

    [ -z "$hush" ] && echo "${app} [${pid}] ${run_msg}"
    return 0
}


ck_start_gxpm()
{
    check_pidfile ${GPM_PIDFILE:?} gxpm && return 0

    echo "Starting GXPM ..."
    local rc=
    ${GPM:?} -q -p ${GPM_PIDFILE} ${GPM_OPTIONS}; rc=$?
    if [ $rc -ne 0 ]; then
        echo >&2 "gxpm failed to start"; return 1
    fi

    sleep 0.5
    check_pidfile ${GPM_PIDFILE:?} gxpm
}


ck_stop_gxpm()
{
    check_pidfile --hush ${GPM_PIDFILE:?} gxpm || return 0

    # NB: check_pidfile sets pid
    echo "Stopping GXPM (${pid:?})"
    if ! kill ${pid}; then
        echo >&2 "Failed to stop GXPM (${pid})"
        return 1
    else
        rm -f ${GPM_PIDFILE:?}
    fi
}


ck_start_channel()
{
    local tag=${1:?}
    local cdir=${GXA_CHANNEL_ROOT:?}/${tag}
    local pidfile=${cdir:?}/vsm.pid

    if [ -f "$cdir/SUSPEND" ]; then
        echo >&2 "<skip> ${tag} (SUSP)"
        return 1
    fi
    if [ -f "$cdir/REPLICA" ]; then
        echo >&2 "<skip> ${tag} (REPL)"
        return 1
    fi

    check_pidfile --hush ${pidfile} ${tag} && return 0

    local logfile=${cdir:?}/vsm.log
    local specfile=${GXA_SPEC_DIR:?}/${tag}.spec

    sudo -u ${GXA_USER:-gigaplus} nohup ${VSM:?} -l ${logfile} ${specfile} >/tmp/nohup.out &

    local wpid=$!
    sleep 0.5
    if ! kill -0 $wpid 2>/dev/null; then
        echo >&2 "${tag}: FAILED to start (pid=$wpid)"
    else
        check_pidfile ${pidfile} ${tag}
        disown $wpid
    fi
}


purge_cha_pidfiles()
{
    local cdir=${1:?}
    local pfile=
    find ${cdir} -name '*.pid' | while read pfile; do
        if check_pidfile --hush ${pfile}; then
            rm ${pfile} && echo >&2 "Removed $pfile"
        fi
    done
}


ck_stop_channel()
{
    local rc=0
    local tag=${1:?}
    local cdir=${GXA_CHANNEL_ROOT:?}/${tag}
    local pidfile=${cdir:?}/vsm.pid

    if check_pidfile --hush ${pidfile} ${tag}; then
        if kill ${pid} 2>/dev/null; then
            echo >&2 "${tag} VSM stopped (pid=$pid)"

            purge_cha_pidfiles ${cdir:?}

            let stopped+=1
            echo "st=${stopped}"
            rc=0
        fi
    else
        echo >&2 "<skip> ${tag}"
        rc=0
    fi

    if ls ${cdir:?}/*.pid 2>/dev/null; then
        echo >&2 "Some (${tag}) apps may be running, checking..."
        sleep 1
        for gpfile in $(ls -1 ${cdir:?}/*.pid 2>/dev/null); do
            if check_pidfile ${gpfile} ${tag}; then
                kill ${pid} || rc=1
                [ $rc -eq 0 ] && echo >&2 "STOPPED app($pid): $gpfile"
            fi
        done
    fi

    return ${rc}
}


ck_start_dwg()
{
    local inum=${1}     # can be empty
    local options=${2}  # can be empty

    local app=${DWG:?}
    local pfile="${GXA_ROOT:?}/run/${app:?}${inum}.pid"
    check_pidfile --hush ${pfile:?} ${app} && return 0

    local start_msg="Starting ${app}#${inum}"
    [ -n "$options" ] && start_msg="${start_msg}: ${options}"

    echo ${start_msg:?}
    ${DWG} -q -p ${pfile:?} ${options}; rc=$?

    sleep 0.5
    check_pidfile ${pfile} ${app}
}


stop_dwags()
{
    local pfile=
    find ${GXA_ROOT:?}/run -type f -name 'dwg*.pid' | \
    while read pfile; do
        if check_pidfile --hush ${pfile} ${DWG}; then
            if ! kill ${pid}; then
                echo >&2 "Failed to stop ${DWG}($pid)"
            else
                rm -f ${pfile}
                echo >&2 "Stopped ${DWG}[$pid]"
            fi
        fi
    done
}


status_dwags()
{
    local pfile=
    local count=0

    find ${GXA_ROOT:?}/run -type f -name 'dwg*.pid' | \
    while read pfile; do
        if check_pidfile "${pfile}" "${DWG}"; then
            let count+=1
        fi
    done

    if [ "$count" -gt 0 ]; then
        echo >&2 "[${DWG}] $count instances"
    else
        echo >&2 "[${DWG}] is not running"
    fi
}


start_dwags()
{
    local count=${1:?}
    local app='dwg'

    local i; for (( i=1; i <= ${count}; i++ )); do
        local opt="${DWG_OPTIONS}"
        [ -n "${PIN_DWG}" ] && opt="${DWG_OPTIONS} --cpu $i"
        ck_start_dwg "$i" "${opt}" || break
    done
}


start_lb()
{
    check_pidfile --hush ${LB_PIDFILE:?} ${FLB:?} && return 0

    echo "Starting ${FLB} ..."
    flb ${LB_OPTIONS} -p "${LB_PIDFILE:?}"; rc=$?
    [ "0" -eq "$rc" ] || return 1

    sleep 0.5
    check_pidfile ${LB_PIDFILE:?} ${FLB:?}
}


stop_lb()
{
    check_pidfile ${LB_PIDFILE:?} ${FLB:?} || return 0

    if ! kill ${pid}; then
        echo >&2 "Failed to stop ${FLB} [$pid]"
    else
        rm -f ${LB_PIDFILE}
        echo >&2 "Stopped ${FLB} [${pid}]"
    fi
}

status_lb()
{
    check_pidfile ${LB_PIDFILE:?} ${FLB:?}
}


start_channels()
{
    local tag=${1:?}
    local cha=
    local rc=$?
    local started=0

    if [ -z "${NOPM}" ]; then
        ck_start_gxpm || return 1
    fi

    if [ "all" = "$tag" ]; then
        find ${GXA_SPEC_DIR} -type f -name '*.spec' | \
            while read sname; do
                cha=$(basename $sname | sed 's/\.spec$//g')
                ck_start_channel ${cha:?}; rc=$?
                [ "$rc" -eq 0 ] && let started+=1
            done
    else
        for cha in `echo ${tag:?}`; do
            ck_start_channel ${cha:?}; rc=$?
            [ "$rc" -eq 0 ] && let started+=1
        done
    fi

    if [ "${started}" -gt 0 ]; then
        echo >&2 "${started} channel(s) started"
        echo >&2 "Pausing for [${CHA_PAUSE}] seconds.."
        sleep ${CHA_PAUSE:?}
    fi

    get_status ${tag}
}

stop_channels()
{
    local tag=${1:?}
    local cha=
    local clist=/tmp/clist.$$

    stopped=0
    echo >&2 "Stopping [${tag}]"
    if [ "all" = "$tag" ]; then
        find ${GXA_SPEC_DIR} -type f -name '*.spec' >${clist}
        while read sname; do
                cha=$(basename $sname | sed 's/\.spec$//g')
                ck_stop_channel ${cha:?}
        done <${clist}; rm -f ${clist}
    else
        for cha in `echo ${tag:?}`; do ck_stop_channel ${cha:?}; done
    fi

    echo >&2 "${stopped} channels stopped."
    if [ "$stopped" -gt 0 ]; then
        echo >&2 "Pausing.." && sleep 1
        get_status ${tag}
    fi
}


shut_down()
{
    stop_channels all
    ck_stop_gxpm
    stop_dwags

    LB_PIDFILE="${GXA_ROOT:?}/run/${FLB:?}.pid"
    stop_lb
}


#
# Main
#

case ${OSNAME} in
    FreeBSD)
        [ -n "${GPX_CONFIG_DIR}" ] || GPX_CONFIG_DIR=/usr/local/etc
        [ -n "${GPX_BIN_DIR}" ] || GPX_BIN_DIR=/usr/local/bin
        ;;
    Linux)
        [ -n "${GPX_CONFIG_DIR}" ] || GPX_CONFIG_DIR=/etc
        [ -n "${GPX_BIN_DIR}" ] || GPX_BIN_DIR=/usr/bin
        ;;
    *) errmsg "${OSNAME} is not supported."
esac

if [ $# -lt 1 ]; then
cat <<_EOT >&2
Usage: $0 [--nopm] [--pause N] start|stop|status|shutdown [{channel}|all]

Options:
    --nopm              = do NOT attempt to start/stop gxpm.
    --pause             = wait N [5] seconds before checking on channels after start.
    --alarm             = generate alarm if channel is offline and/or oversized.
    --dwg[pin] {num}    = start/stop/status N dwg instances [pin to CPU cores].
    --lb                = start/stop/status load balancer.
    --debug mask        = list app(s) in debug mode (lb:dwg:pm)
Note:
    shutdown mode needs no parameters, it will stop all channels and modules.

GigA+ channel launch script.
This script is part of GigA+. Copyright 2017 by Pavel Cherenkov
_EOT
    exit 1
fi


envfile="${GPX_CONFIG_DIR}/gigaplus.env"
[ -f "./gigaplus.env" ] && envfile="./gigaplus.env"

if [ -f "${envfile:?}" ]; then
    . ${envfile}
    : ${GXA_ROOT:?}
    : ${GXA_CHANNEL_ROOT:?}
else
    errmsg "Missing ${envfile}"
fi

if [ -n "${GPX_BIN_DIR}" ]; then
    GPM=${GPX_BIN_DIR}/gxpm
    VSM=${GPX_BIN_DIR}/vsm
    DWG=${GPX_BIN_DIR}/dwg
    FLB=${GPX_BIN_DIR}/flb
else
    : ${GPM:=gxpm}
    : ${VSM:=vsm}
    : ${DWG:=dwg}
    : ${FLB:=flb}
fi

if type ${MAIL_AGENT:?} >/dev/null 2>&1; then
    can_email=1
fi

GPM_PIDFILE="${GXA_ROOT:?}/run/gxpm.pid"

[ -n "${GXA_SPEC_DIR}" ] || GXA_SPEC_DIR="${GXA_ROOT:?}/etc"

while [ -n "$1" ]; do
    if [ '--nopm' = "$1" ]; then
        NOPM=1; shift
        continue
    fi

    if [ '--pause' = "$1" ]; then
        CHA_PAUSE=$2; shift 2
        continue
    else
        CHA_PAUSE=5
    fi

    if [ "--alert" = "$1" ]; then
        MUST_ALERT=1; shift
        continue
    fi

    if [ "--dwg" = "$1" ]; then
        USE_DWAGS=${2:?}; PIN_DWG=;
        if ! expr ${USE_DWAGS} + 1 >/dev/null 2>&1; then
            echo >&2 "Invalid --dwg parameter"; exit 1
        fi
        shift 2; continue
    fi

    if [ "--dwgpin" = "$1" ]; then
        USE_DWAGS=${2:?}; PIN_DWG=1;
        if ! expr ${USE_DWAGS} + 1 >/dev/null 2>&1; then
            echo >&2 "Invalid --dwgpin parameter"; exit 1
        fi
        shift 2; continue
    fi

    if [ "--lb" = "$1" ]; then
        LB_PIDFILE="${GXA_ROOT:?}/run/${FLB:?}.pid"
        shift; continue
    fi

    if [ "--debug" = "$1" ]; then
        dmask=${2}
        case "$dmask" in
            *lb*)   LB_OPTIONS="-L debug ${LB_OPTIONS}";
                    echo >&2 "[lb-debug]"
                    ;;
        esac
        case "$dmask" in
            *dwg*)  DWG_OPTIONS="-L debug ${DWG_OPTIONS}"
                    echo >&2 "[dwg-debug]"
                    ;;
        esac
        case "$dmask" in
            *pm*)   GPM_OPTIONS="-L debug ${GPM_OPTIONS}"
                    echo "[gpxm-debug]"
                    ;;
        esac
        shift 2; continue
    fi

    break
done

cmd=${1}
if [ -z "$cmd" ]; then
    echo >&2 "Missing command (start, stop, etc.)"
    exit 1
fi

shift
cha_tag="$@"
[ -n "${cha_tag}" ] || cha_tag='all'

euid=$(id -u)
if [ "status" != "$cmd" ]; then
    if [ $euid -ne 0 ]; then
        echo >&2 "Must be root to ${cmd} channel(s)"
        exit 1
    fi
fi

case "$cmd" in
    'start')
        if [ -n "${USE_DWAGS}" ]; then
            start_dwags "${USE_DWAGS}" || return 1
        fi
        if [ -n "${LB_PIDFILE}" ]; then
            start_lb || exit 1
        fi
        [ "-" = "${cha_tag}" ] || start_channels "${cha_tag}"
        ;;
    'stop')
        [ "-" = "${cha_tag}" ] || stop_channels "${cha_tag}"
        [ -n "${USE_DWAGS}" ] && stop_dwags
        [ -n "${LB_PIDFILE}" ] && stop_lb
        ;;
    'status')
        [ -n "${USE_DWAGS}" ] && status_dwags
        [ -n "${LB_PIDFILE}" ] && status_lb
        [ "-" = "${cha_tag}" ] || get_status "${cha_tag}"
        ;;
    'shutdown') shut_down ;;
    # 'fmtu') print_fmt_uptime ${2:?} ;;
    *) echo >&2 "Unrecognized mode: $cmd"; exit 2 ;;
esac

# __EOF__

