#!/usr/bin/env bash
# @(#) Launch script for GigA+ gxws/gxng.
#
# Example: sudo GXA_PFX=/opt/gxa GWS_OPTIONS=-vvv GNG_OPTIONS=-vvv ./gxa-requests start 3
#
#*****************************************************************
# Copyright 2012-2017 Pavel V. Cherenkov (pcherenkov@gigapxy.com)
#
#  This file is part of GigA+.
#
#  GigA+ is commercial software: all rights to it belong
#  to Pavel V. Cherenkov, its author.
#
#*****************************************************************/

errmsg()
{
  echo >&2 $@; exit 1
}


#############################
# Variables
#############################

: ${GXA_PFX:='/opt/gxa'}

RM=/bin/rm
MKDIR=/bin/mkdir
CAT=/bin/cat
OSNAME=`uname -s`

# NB: set GXA_PFX to run in a sandbox.

GPX_PIDFILE_DIR=${GXA_PFX}/run
GPX_LOG_DIR=${GXA_PFX}/log
GPX_BUF_DIR=${GXA_PFX}/buf

case ${OSNAME} in
    FreeBSD)
        [ -n "${GPX_CONFIG_DIR}" ] || GPX_CONFIG_DIR=/usr/local/etc
        [ -n "${GPX_BIN_DIR}" ] || GPX_BIN_DIR=/usr/local/bin
        ;;
    Linux)
        [ -n "${GPX_CONFIG_DIR}" ] || GPX_CONFIG_DIR=/etc
        [ -n "${GPX_BIN_DIR}" ] || GPX_BIN_DIR=/usr/bin
        ;;
    *) errmsg "${OSNAME} is not supported."
esac

envdir=.
[ -f ${envdir}/gigaplus.env ] || envdir=${GPX_CONFIG_DIR}
envfile="${envdir}/gigaplus.env"

if [ -f "${envfile}" ]; then
    . ${envfile}
else
    errmsg "Missing ${envfile}"
fi

if [ -n "${GPX_BIN_DIR}" ]; then
    GWS=${GPX_BIN_DIR}/gxws
    GNG=${GPX_BIN_DIR}/gxng
    GPM=${GPX_BIN_DIR}/gxpm
else
    : ${GWS:=gxws}
    : ${GNG:=gxng}
    : ${GPM:=gxpm}
fi

[ -z "${GXA_CONFIG}" ] && GXA_CONFIG=${GPX_CONFIG_DIR}/gigaplus.conf

GXWS_PIDFILE=${GPX_PIDFILE_DIR}/gxws.pid
# The template expands to gng1.pid, gng2.pid, etc.
GXNG_PIDFILE_TPL=${GPX_PIDFILE_DIR}/gxngXX.pid

GXPM_PIDFILE=${GPX_PIDFILE_DIR}/gxpm.pid

GXWS_LOG=${GPX_LOG_DIR}/gxws.log
# The template expands to gng1.log, gng2.log, etc.
GXNG_LOG_TPL=${GPX_LOG_DIR}/gxngXX.log

# Specify any additional command-line options.
GWS_OPTIONS="${GWS_OPTIONS}"
GNG_OPTIONS="${GNG_OPTIONS}"

# Seconds to wait for components to start up.
: ${WAIT_SEC:='1'}


#############################
# Functions
#############################

usage()
{
${CAT} <<__EOM
Usage: $1 OPTIONS start|stop|shutdown|status {num|none} [nobuf-ngs]
OPTIONS:
    --nopm:         do not launch gxpm;
    --killpm:       will stop gxpm (in stop mode);
    --nopin:        no CPU affinity to be set;
    --debug mask    list app(s) in debug mode (ws:ng:pm)
PARAMETERS:
    num       = number of (gxng) engines to launch [1..64] or 'none'
    nobuf-ngs = additional engines to serve files/segments

Set GXA_DEBUG=1 to see what commands are run.
Example: $1 --nopm start 3

$1: GigA+ gxpm/gxws/gxng launch script (feel free to modify).
$1 is part of GigA+. GigA+ is Copyright (C) 2016 - 2018 Pavel Cherenkov

__EOM
  exit 1
}


check_prerequisites()
{
  type ${GWS:?} >/dev/null && type ${GNG:?} >/dev/null || \
    errmsg "Missing gxws=[${GWS}], gxng=[${GNG}]"

  type ${GPM:?} >/dev/null || errmsg "Missing gxpm=${GPM}"

  if [ ! -d "${GPX_CONFIG_DIR}" ]; then
    errmsg "$0: config directory [${GPX_CONFIG_DIR}] does not exist."
  fi

  if [ ! -d "${GPX_LOG_DIR}" ]; then
    ${MKDIR} -p ${GPX_LOG_DIR} || errmsg "$0: failed to mkdir [${GPX_LOG_DIR}]"
  fi
}

check_pidfile()
{
    local hush=

    if [ '--hush' = "$1" ]; then
        hush=1; shift
    fi

    local fname=$1
    local app=${2:-'process'}
    local run_msg=${3:-'is running'}

    if [ -f "${fname}" ]; then
        pid=$(cat ${fname})
        [ -n "$pid" ] || return 1

        if ! ps -p ${pid} 2>/dev/null >&2; then
            [ -z "$hush" ] && echo "Stale pidfile [${fname}] found; removing."
            rm -f ${fname}
        fi
    fi

    if [ ! -f "${fname}" ]; then
        [ -z "$hush" ] && echo "[${app}] is not running."
        return 1
    fi

    [ -z "$hush" ] && echo "${app} [${pid}] ${run_msg}"
    return 0
}

gxy_start()
{
    check_pidfile ${GXWS_PIDFILE} 'GXWS' && return 1

    GNG_BUF=${1:-'1'}
    GNG_FILE=${2:-'0'}
    GNG_NUM=$(( ${GNG_BUF} + ${GNG_FILE} ))

    # GWS has to be launched first.
    if [ -n "${GXA_DEBUG}" ]; then
        echo ${GWS} ${GWS_OPTIONS} -q -l ${GXWS_LOG} -p ${GXWS_PIDFILE}
    fi

    if [ -z "${NO_GPM}" ]; then
        if ! check_pidfile ${GXPM_PIDFILE} 'GXPM'; then
            if [ -n "${GXA_DEBUG}" ]; then
                echo ${GPM} -q ${GPM_OPTIONS}
            fi

            echo "Starting GXPM ..."
            ${GPM} -q ${GPM_OPTIONS} || errmsg "Failed to start GXPM"
        fi
    fi

    echo "Starting GXWS ..."
    ${GWS} -q ${GWS_OPTIONS} -l ${GXWS_LOG} -p ${GXWS_PIDFILE}

    i=1
    while [ "$i" -le "${GNG_NUM:?}" ]; do
        pidfile=`echo ${GXNG_PIDFILE_TPL} | sed "s/XX/$i/g"`
        logfile=`echo ${GXNG_LOG_TPL} | sed "s/XX/$i/g"`

        cpu_affinity=
        if [ -z "${NOPIN}" ]; then
            cpunum=`expr $i - 1`
            cpu_affinity="-P ${cpunum}"
        fi

        gmode='regular'
        nobuf=
        if [ "$i" -gt "${GNG_BUF}" ]; then
            nobuf="--nobufd"
            gmode="file"
        fi

        if [ -n "${GXA_DEBUG}" ]; then
            echo ${GNG} -q ${GNG_OPTIONS} -l ${logfile} -p ${pidfile} ${cpu_affinity} ${nobuf}
        fi

        ${GNG} -q ${GNG_OPTIONS} -l ${logfile} -p ${pidfile} ${cpu_affinity} ${nobuf}
        echo "Starting (${gmode}) GXNG${i} ..."

        i=`expr $i + 1`
    done

    echo "Pausing for ${WAIT_SEC} second(s)."
    sleep ${WAIT_SEC}

    gws_pid=`${CAT} ${GXWS_PIDFILE} 2>/dev/null`
    if [ $? -eq 0 ] && kill -0 ${gws_pid}; then
        echo "GXWS [${gws_pid}] STARTED"
    else
        errmsg "GXWS [${gws_pid}] FAILED, see ${GXWS_LOG}"
    fi

    ng_err=0; i=1
    while [ "$i" -le "${GNG_NUM}" ]; do
        pidfile=`echo ${GXNG_PIDFILE_TPL} | sed "s/XX/$i/g"`
        logfile=`echo ${GXNG_LOG_TPL} | sed "s/XX/$i/g"`

        gng_pid=`$CAT ${pidfile}`
        if kill -0 ${gng_pid} 2>/dev/null; then
            echo "GXNG [${gng_pid}] STARTED"
        else
            echo "GXNG [${gng_pid}] FAILED, see ${logfile}"
            ng_err=1
        fi

        i=`expr $i + 1`
    done

    check_pidfile ${GXPM_PIDFILE} 'GXPM' STARTED

    if [ "$ng_err" -gt 0 ]; then
        echo "$0: some GNG's could not start, shutting down GWS."
        kill ${gws_pid}
    fi
}

gxy_status()
{
    [ -z "${NO_GPM}" ] && check_pidfile ${GXPM_PIDFILE} 'GXPM'
    check_pidfile ${GXWS_PIDFILE} 'GXWS' || return 1

    ls ${GPX_PIDFILE_DIR}/gxng*.pid 2>/dev/null | while read fname; do
        check_pidfile ${fname} 'GXNG'
    done
}


gxy_stop()
{
    if [ -n "${KILL_GPM}" ]; then
        if check_pidfile ${GXPM_PIDFILE} 'GXPM'; then
            if kill ${pid}; then
                echo "GXPM [${pid}] stopped."
            else
                echo "GXPM [${pid}] FAILED to stop."
            fi
        fi
    fi

    check_pidfile ${GXWS_PIDFILE} 'GXWS' || return 1
    if kill ${pid}; then
        echo "GXWS [${pid}] stopped."
    else
        echo "GXWS [${pid}] FAILED to stop."
    fi

    # if ws.ng.force_shutdown is enabled, GNG's should quit.
    sleep ${WAIT_SEC}

    ls ${GPX_PIDFILE_DIR}/gxng*.pid 2>/dev/null | while read fname; do
        check_pidfile ${fname} 'GXNG' || return 1
        if kill ${pid}; then
            echo "GXNG [${pid}] stopped."
        else
            echo "GXNG [${pid}] FAILED to stop."
        fi
    done
}


#############################
# Main
#############################

[ $# -lt 1 ] && usage $0

while true; do
    if [ '--nopm' = "$1" ]; then
        NO_GPM=1; shift
        continue
    fi
    if [ '--killpm' = "$1" ]; then
        KILL_GPM=1; shift
        continue
    fi
    if [ '--nopin' = "$1" ]; then
        NOPIN=1; shift
        continue
    fi

    if [ "--debug" = "$1" ]; then
        dmask=${2}
        case "$dmask" in
            *ws*)   GWS_OPTIONS="-vvv ${GWS_OPTIONS}";
                    echo >&2 "[gxws-debug]"
                    ;;
        esac
        case "$dmask" in
            *ng*)  GNG_OPTIONS="-vvv ${GNG_OPTIONS}"
                    echo >&2 "[gxng-debug]"
                    ;;
        esac
        case "$dmask" in
            *pm*)   GPM_OPTIONS="-L debug ${GPM_OPTIONS}"
                    echo "[gxpm-debug]"
                    ;;
        esac
        shift 2; continue
    fi
    break
done

check_prerequisites

case "$1" in
    'start'|'START')
        if [ $# -ge 2 ]; then
            if [ "none" = "$2" ]; then
                ngnum=0; nobuf_ngs=0
            else
                # Number of (gng) engines to launch.
                ngnum=`expr $2 + 0 2>/dev/null`
                [ "0" != "$2" -a $? -ne 0 ] && errmsg "$0: invalid parameter format."

                if [ $ngnum -lt 0 -o ${ngnum} -gt 64 ]; then
                    errmsg "num [${ngnum}] is out of range."
                fi

                nobuf_ngs=0
                [ -n "$3" ] && nobuf_ngs=`expr "$3 + 0" 2>/dev/null`
            fi
        else
            errmsg "Missing mandatory parameter num = number of engines"
        fi
        gxy_start ${ngnum} ${nobuf_ngs}
        ;;

    'stop'|'STOP')
        gxy_stop
        ;;
    'shutdown'|'SHUTDOWN')
        KILL_GPM=1; gxy_stop
        ;;

    'status'|'STATUS')
        gxy_status
        ;;
    *) errmsg "Unrecognized command: $1" ;;
esac


# __EOF__

